#! /usr/bin/python -u
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
# Copyright 2012-2016 Cumulus Networks LLC, all rights reserved

# This program is a Net-SNMP pass persist script that services the
# Resource Query MIB that's been defined by Cumulus Networks.

import json
import subprocess
import syslog
import getopt
import sys
import os.path
import re
try:
    from collections import OrderedDict
except ImportError:
    import ordereddict

import snmp_passpersist as snmp

oid_base = '.1.3.6.1.4.1.40310.1'
header_regexp = 'Timestamp:\s*([\w,:, ]+)\s*Poll Interval: ([0-9]+) msec Measurement Interval: ([0-9])+'
# To run from the commandline for debugging,
# simply set debug to non-zero and run to see
# STDOUT output.
debug = 0

def pp_add_int(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_int(oid, val)

def pp_add_str(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_str(oid, val)

def pp_add_cnt64(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_cnt_64bit(oid, val)

def get_json_output(commandList=None):
    '''
    This grabs the JSON output of a CLI command and returns an array.
    '''
    global syslog_already_logged
    outArray = {}
    if (not commandList) or (commandList == []):
        syslog.syslog('Error: called get_json_output with'
                      ' invalid command=%s' % \
                      commandList)
        syslog_already_logged = True
        return {}
    try:
        outArray = json.loads(subprocess.check_output((commandList),
                                                      shell=False),
                              encoding="latin-1")
    except Exception as e:
        outArray = {}
        syslog.syslog('Error: command %s EXCEPTION=%s' % \
                      (' '.join(commandList), e))
        syslog_already_logged = True
    return outArray

def update_mlx_buf_utilizn():
    """
    Extract the pool buffer stats for mellanox platform from mlxcmd
    """
    try:
        poolStats = get_json_output(commandList=['sudo', '/usr/lib/cumulus/mlxcmd', '--json', 'buffers', 'pool', 'counters', '--cells'])
        buffStats = get_json_output(commandList=['sudo', '/usr/lib/cumulus/mlxcmd', '--json', 'buffers', 'port_reserved', 'counters', '--cells', '--port', 'swp*'])
    except Exception as e:
        # it is ok to return
        return

    bufUtilTbl = "3.4"              # .1.3.6.1.4.1.40310.1.3.4
    bufServicePoolID = "1.1"        #                         .1.1
    bufMax = "1.3"                  #                         .1.3
    bufCurr = "1.7"                 #                         .1.7

    portPgbufUtil = "3.5"           # .1.3.6.1.4.1.40310.1.3.5
    portName = "1.1"                #                         .1.1

    portPoolbufUtil = "3.6"         #                       .6
    portTcbufUtil = "3.7"           #                       .7
    bufCellSize = "3.8"             #                       .8
    cellSize = 0

    for poolId in poolStats:
        poolStat = poolStats[poolId]
        currUsage = poolStat['current']
        maxUsage = poolStat['watermark']
        cellSize = poolStat['cell_unit_size']
        pool_id = poolStat['id']
        pp_add_int("%s.%s.%s"%(bufUtilTbl, bufServicePoolID, pool_id), int(pool_id))
        pp_add_int("%s.%s.%s"%(bufUtilTbl, bufMax, pool_id), int(maxUsage))
        pp_add_int("%s.%s.%s"%(bufUtilTbl, bufCurr, pool_id), int(currUsage))

    idx = 0
    for portBufStat in buffStats:
        if portBufStat['port'] == 'MC_PORT':
            continue
        port_name = portBufStat['port']
        idx += 1
        pp_add_str("%s.%s.%d"%(portPgbufUtil,portName,idx), port_name)
        pp_add_str("%s.%s.%d"%(portPoolbufUtil,portName,idx), port_name)
        pp_add_str("%s.%s.%d"%(portTcbufUtil,portName,idx), port_name)

        for stat in portBufStat['counter_list']:
            if stat['type'] == 'INGRESS_PORT_PRIORITY_GROUP':
                pg_id = stat['pg_id']
                portPgCurrUsage = '1.{}'.format(2*int(pg_id) + 2)       # .1.3.6.1.4.1.40310.1.3.5.1.[2,4,6,8..]
                portPgMaxUsage = '1.{}'.format(2*int(pg_id) + 3)        # .1.3.6.1.4.1.40310.1.3.5.1.[3,5,7,9..]
                currUsage = stat['current']
                maxUsage = stat['watermark']
                pp_add_cnt64("%s.%s.%s"%(portPgbufUtil, portPgCurrUsage, idx), int(currUsage))
                pp_add_cnt64("%s.%s.%s"%(portPgbufUtil, portPgMaxUsage, idx), int(maxUsage))

            if stat['type'] == 'EGRESS_PORT_TRAFFIC_CLASS':
                tc_id = stat['tclass']
                portTcCurrUsage = '1.{}'.format(2*int(tc_id) + 2)       # .1.3.6.1.4.1.40310.1.3.7.1.[2,4,6,8..]
                portTcMaxUsage = '1.{}'.format(2*int(tc_id) + 3)        # .1.3.6.1.4.1.40310.1.3.7.1.[3,5,7,9..]
                currUsage = stat['current']
                maxUsage = stat['watermark']
                pp_add_cnt64("%s.%s.%s"%(portTcbufUtil, portTcCurrUsage, idx), int(currUsage))
                pp_add_cnt64("%s.%s.%s"%(portTcbufUtil, portTcMaxUsage, idx), int(maxUsage))

            if stat['type'] == 'INGRESS_PORT':
                pool_id = stat['pool_id']
                portPoolCurrUsage = '1.{}'.format(2*int(pool_id) + 2)       # .1.3.6.1.4.1.40310.1.3.6.1.[2,4,6,8..]
                portPoolMaxUsage = '1.{}'.format(2*int(pool_id) + 3)        # .1.3.6.1.4.1.40310.1.3.6.1.[3,5,7,9..]
                currUsage = stat['current']
                maxUsage = stat['watermark']
                pp_add_cnt64("%s.%s.%s"%(portPoolbufUtil, portPoolCurrUsage, idx), int(currUsage))
                pp_add_cnt64("%s.%s.%s"%(portPoolbufUtil, portPoolMaxUsage, idx), int(maxUsage))

            if stat['type'] == 'EGRESS_PORT':
                pool_id = stat['pool_id']
                portPoolCurrUsage = '1.{}'.format(2*int(pool_id))           # .1.3.6.1.4.1.40310.1.3.6.1.[22,24,26,28..]
                portPoolMaxUsage = '1.{}'.format(2*int(pool_id) + 1)        # .1.3.6.1.4.1.40310.1.3.6.1.[23,25,27,29..]
                currUsage = stat['current']
                maxUsage = stat['watermark']
                pp_add_cnt64("%s.%s.%s"%(portPoolbufUtil, portPoolCurrUsage, idx), int(currUsage))
                pp_add_cnt64("%s.%s.%s"%(portPoolbufUtil, portPoolMaxUsage, idx), int(maxUsage))

    pp_add_int("%s"%(bufCellSize), int(cellSize))

def update_buf_utilizn():
    """
    Extract the buffer utilization stats from the file and stuff it into SNMP
    The file /tmp/hal-buf-util.txt will get created only if user does:
        echo "1" > /cumulus/switchd/config/buf_util/measure_interval
        echo "1" > /cumulus/switchd/config/buf_util/poll_interval
    """
    try:
        with open('/tmp/hal-buf-util.txt') as x: buf_lines = x.readlines()
    except Exception as e:
        # this is not important as they may not have turned these stats on
        return


    # This is based on the assumption that the format looks like this:
    # Timestamp: Wed Jul 25 07:22:10 2012 Poll Interval: 10 msec Measurement
    # Interval: 1 min(s) Next 4 lines are for each service pool: poolID, min,
    # max, avg, variance, SD
    # 0, 0, 0, 0.0000    , 0.0000    , 0.0000
    # 1, 0, 0, 0.0000    , 0.0000    , 0.0000
    # 2, 0, 0, 0.0000    , 0.0000    , 0.0000
    # 3, 0, 0, 0.0000    , 0.0000    , 0.0000

    # Extract the non-table entries from the first line
    m = re.match(header_regexp, buf_lines[0])
    if (m != None):
        tstamp = m.group(1).strip()
        poll_val = m.group(2)
        meas_val = m.group(3)
    else:
        tstamp = poll_val = meas_val = -1

    # Convert timestamp (ctime format) into what SNMP TimeStamp format demands.
    # Can't really support the TimeStamp format via the pass_persist mechanism.
    # So, we'll just convert it into the format of TimeStamp and let the client
    # deal with it.
    # For crying out loud, the time module in Cumulus doesn't support strptime()
    # So do the work here. Luckily, SNMP only needs the month converted
    MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
              "Sep", "Oct", "Nov", "Dec"]

    ts = tstamp.split()
    if ts[1] in MONTHS:
        mnth = MONTHS.index(ts[1])+1

    tstamp = "%s-%d-%s,%s.0"%(ts[4], mnth, ts[2], ts[3])

    pp_add_str("3.1", tstamp)
    pp_add_int("3.2", poll_val)
    pp_add_int("3.3", meas_val)

    # Lets parse the rest of the lines now for the actual buffer data
    baseoid = "3.4.1"
    try:
        s = OrderedDict()
    except Exception as e:
        syslog.syslog('resq_pp:  EXCEPTION=%s' % e)
        s = ordereddict.OrderedDict()

    for line in buf_lines[2:]:
        if line.strip():
            l = ([ i.strip() for i in line.split(",") ])
            s.update(dict({l[0]:l[1:]}))

    for idx in s.iterkeys():
        pp_add_int("%s.%d.%s"%(baseoid, 1, idx), int(idx))

    for i in range(0,2):
        for idx, vals in s.iteritems():
            pp_add_int("%s.%d.%s"%(baseoid, i+2, idx), vals[i])

    for i in range(2,5):
        for idx, vals in s.iteritems():
            pp_add_str("%s.%d.%s"%(baseoid, i+2, idx), vals[i])


def get_resourceinfo():
    try:
        reslines = subprocess.check_output((['/usr/cumulus/bin/cl-resource-query','-k']),
                                           stderr=subprocess.STDOUT,
                                           shell=False).splitlines()
    except Exception as e:
        # we need to log the error if we can't get resources
        syslog.syslog("resq_pp:  Exception=%s" % e)

    if debug:
        print reslines
    if not reslines:
        return

    """parses output from cl-resource-query -k
            host_entry_count=2
            host_entry_max=8192
            host_v4_entry_count=2
            host_v6_entry_count=0
            route_0_entry_count=5
            route_0_entry_max=32668
            route_1_entry_count=13
            route_1_entry_max=16384
            route_v4_entry_count=5
            route_v6_entry_count=13
            route_total_entry_count=18
            route_total_entry_max=32768
            ecmp_nh_entry_count=0
            ecmp_nh_entry_max=16327
            mac_entry_count=0
            mac_entry_max=32768
            in_acl_entry_count=60
            in_acl_entry_max=1280
            in_acl_counter_count=81
            in_acl_counter_max=1280
            in_acl_meter_count=21
            in_acl_meter_max=4096
            in_acl_slice_count=2
            in_acl_slice_max=4
            eg_acl_entry_count=24
            eg_acl_entry_max=256
            eg_acl_counter_count=48
            eg_acl_counter_max=1024
            eg_acl_meter_count=24
            eg_acl_meter_max=256
            eg_acl_slice_count=1
            eg_acl_slice_max=1


    And returns dictionary in the following format:
            {'eg_acl_slice_count': 1, 'host_entry_count': 2, 'eg_acl_counter_max': 1024, 'in_acl_slice_count': 2, 'host_entry_max': 8192, 'in_acl_meter_count': 21, 'route_total_entry_count': 18, 'mac_entry_count': 0, 'eg_acl_meter_max': 256, 'route_v4_entry_count': 5, 'host_v6_entry_count': 0, 'ecmp_nh_entry_count': 0, 'in_acl_slice_max': 4, 'eg_acl_counter_count': 48, 'ecmp_nh_entry_max': 16327, 'route_1_entry_count': 13, 'in_acl_counter_count': 81, 'route_total_entry_max': 32768, 'in_acl_counter_max': 1280, 'host_v4_entry_count': 2, 'in_acl_meter_max': 4096, 'eg_acl_entry_max': 256, 'route_0_entry_max': 32668, 'in_acl_entry_count': 60, 'in_acl_entry_max': 1280, 'route_v6_entry_count': 13, 'route_1_entry_max': 16384, 'eg_acl_entry_count': 24, 'route_0_entry_count': 5, 'eg_acl_slice_max': 1, 'mac_entry_max': 32768, 'eg_acl_meter_count': 24}
    """

    resdict = {}
    for line in reslines:
        if '=' in line:
            (rname, rvalstr) = line.split('=')
        else:
            continue
        try:
            resdict[rname.strip()] = int(rvalstr.strip())
        except Exception as e:
            syslog.syslog('resq_pp:  EXCEPTION=%s' % e)

    return resdict

def update_resourceinfo():
    resdict = get_resourceinfo()
    if debug:
        print('resdict=%s'%resdict)
    # L3 entries
    try:
        pp_add_int('1.1', resdict.get('host_0_entry_count',0))
        pp_add_int('1.2', resdict.get('host_0_entry_max',0))

        pp_add_int('1.3', resdict.get('route_total_entry_count',0))
        pp_add_int('1.4', resdict.get('route_total_entry_max',0))

        pp_add_int('1.5', resdict.get('route_total_entry_count',0))
        pp_add_int('1.6', resdict.get('route_total_entry_max',0))

        pp_add_int('1.9', resdict.get('ecmp_nh_entry_count',0))
        pp_add_int('1.10', resdict.get('ecmp_nh_entry_max',0))

        pp_add_int('1.11', resdict.get('in_acl_entry_count',0))
        pp_add_int('1.12', resdict.get('in_acl_entry_max',0))
        pp_add_int('1.13', resdict.get('in_acl_counter_count',0))
        pp_add_int('1.14', resdict.get('in_acl_counter_max',0))
        pp_add_int('1.15', resdict.get('in_acl_meter_count',0))
        pp_add_int('1.16', resdict.get('in_acl_meter_max',0))
        pp_add_int('1.17', resdict.get('in_acl_slice_count',0))
        pp_add_int('1.18', resdict.get('in_acl_slice_max',0))
        pp_add_int('1.19', resdict.get('eg_acl_entry_count',0))
        pp_add_int('1.20', resdict.get('eg_acl_entry_max',0))
        pp_add_int('1.21', resdict.get('eg_acl_counter_count',0))
        pp_add_int('1.22', resdict.get('eg_acl_counter_max',0))
        pp_add_int('1.23', resdict.get('eg_acl_meter_count',0))
        pp_add_int('1.24', resdict.get('eg_acl_meter_max',0))
        pp_add_int('1.25', resdict.get('eg_acl_slice_count',0))
        pp_add_int('1.26', resdict.get('eg_acl_slice_max',0))

        # L2 entries
        pp_add_int('2.1', resdict.get('mac_entry_count',0))
        pp_add_int('2.2', resdict.get('mac_entry_max',0))
    except Exception as e:
        syslog.syslog('resq_pp: ethtool EXCEPTION=%s' % e)

def update():
    update_resourceinfo()
    update_buf_utilizn()
    update_mlx_buf_utilizn()

if debug:
    update()
else:
    pp = snmp.PassPersist(oid_base)
    pp.start(update, 37)           # So we won't ever conflict with the writer

